;; Generated package description from perl-doc.el  -*- no-byte-compile: t -*-
(define-package "perl-doc" "0.1.0.20220807.302" "Read Perl documentation" '((emacs "27")) :commit "deccc5790ab5026460550a728f9552604b67194c" :authors '(("Harald Jörg" . "haj@posteo.de")) :maintainer '("Harald Jörg" . "haj@posteo.de") :keywords '("languages") :url "https://github.com/HaraldJoerg/perl-doc")
